/*
 * QEMU LOONGARCH timer support
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/loongarch/cpudevs.h"
#include "qemu/timer.h"
#include "sysemu/kvm.h"
#include "internal.h"
#include "hw/irq.h"


#ifdef DEBUG_TIMER
#define debug_timer(fmt, args...) printf("%s(%d)-%s -> " #fmt "\n", \
                                 __FILE__, __LINE__, __func__, ##args);
#else
#define debug_timer(fmt, args...)
#endif

#define TIMER_PERIOD                10 /* 10 ns period for 100 Mhz frequency */
#define STABLETIMER_TICK_MASK       0xfffffffffffcUL
#define STABLETIMER_ENABLE          0x1UL
#define STABLETIMER_PERIOD          0x2UL

/* return random value in [low, high] */
uint32_t cpu_loongarch_get_random_ls3a5k_tlb(uint32_t low, uint32_t high)
{
    static uint32_t seed = 5;
    static uint32_t prev_idx;
    uint32_t idx;
    uint32_t nb_rand_tlb = high - low + 1;

    do {
        seed = 1103515245 * seed + 12345;
        idx = (seed >> 16) % nb_rand_tlb + low;
    } while (idx == prev_idx);
    prev_idx = idx;

    return idx;
}

/* LOONGARCH timer */
uint64_t cpu_loongarch_get_stable_counter(CPULOONGARCHState *env)
{
    return qemu_clock_get_ns(QEMU_CLOCK_VIRTUAL) / TIMER_PERIOD;
}

uint64_t cpu_loongarch_get_stable_timer_ticks(CPULOONGARCHState *env)
{
    uint64_t now, expire;

    now = qemu_clock_get_ns(QEMU_CLOCK_VIRTUAL);
    expire = timer_expire_time_ns(env->timer);

    return (expire - now) / TIMER_PERIOD;
}

void cpu_loongarch_store_stable_timer_config(CPULOONGARCHState *env,
                                             uint64_t value)
{
    uint64_t now, next;

    env->CSR_TCFG = value;
    if (value & STABLETIMER_ENABLE) {
        now = qemu_clock_get_ns(QEMU_CLOCK_VIRTUAL);
        next = now + (value & STABLETIMER_TICK_MASK) * TIMER_PERIOD;
        timer_mod(env->timer, next);
    }
    debug_timer("0x%lx 0x%lx now 0x%lx, next 0x%lx",
                value, env->CSR_TCFG, now, next);
}

static void loongarch_stable_timer_cb(void *opaque)
{
    CPULOONGARCHState *env;
    uint64_t now, next;

    env = opaque;
    debug_timer();
    if (env->CSR_TCFG & STABLETIMER_PERIOD) {
        now = qemu_clock_get_ns(QEMU_CLOCK_VIRTUAL);
        next = now + (env->CSR_TCFG & STABLETIMER_TICK_MASK) * TIMER_PERIOD;
        timer_mod(env->timer, next);
    } else {
        env->CSR_TCFG &= ~STABLETIMER_ENABLE;
    }

   qemu_irq_raise(env->irq[IRQ_TIMER]);

}

void cpu_loongarch_clock_init(LOONGARCHCPU *cpu)
{
    CPULOONGARCHState *env = &cpu->env;

    /*
     * If we're in KVM mode, don't create the periodic timer, that is handled in
     * kernel.
     */
    if (!kvm_enabled()) {
        env->timer = timer_new_ns(QEMU_CLOCK_VIRTUAL,
                                  &loongarch_stable_timer_cb, env);
    }
}
